/*******************************************************************************
 * Copyright (c) 2005 BEA Systems, Inc. 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *******************************************************************************/

package org.example.processor.internal.webservice;

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.sun.mirror.apt.AnnotationProcessor;
import com.sun.mirror.apt.AnnotationProcessorEnvironment;
import com.sun.mirror.apt.Messager;
import com.sun.mirror.declaration.AnnotationMirror;
import com.sun.mirror.declaration.AnnotationTypeDeclaration;
import com.sun.mirror.declaration.AnnotationTypeElementDeclaration;
import com.sun.mirror.declaration.AnnotationValue;
import com.sun.mirror.declaration.ClassDeclaration;
import com.sun.mirror.declaration.Declaration;
import com.sun.mirror.declaration.FieldDeclaration;
import com.sun.mirror.declaration.MethodDeclaration;
import com.sun.mirror.declaration.ParameterDeclaration;
import com.sun.mirror.type.ClassType;
import com.sun.mirror.type.PrimitiveType;
import com.sun.mirror.type.TypeMirror;
import com.sun.mirror.util.DeclarationFilter;

public class Processor implements AnnotationProcessor {
// TODO: assignment 2: 
//	examine operationName element for valid value, e.g. 
//		@WebMethod(operationName="???")
//	Examine each complex parameter type for getters, e.g. 
//		@WebMethod public void foo(ComplexType ct) {...}
	
	public static final String annotationNameWebService = "javax.jws.WebService";
	public static final String annotationNameWebMethod = "javax.jws.WebMethod";
	public static final String elementNameOperationName = "operationName";

	
	
	private final Set<AnnotationTypeDeclaration> _atds;
	private final AnnotationProcessorEnvironment _env;
	private final Pattern alphaMatch = Pattern.compile("[a-zA-Z]");

	public Processor(Set<AnnotationTypeDeclaration> atds,
			AnnotationProcessorEnvironment env) {
		_atds = atds;
		_env = env;
		if (APFactory.TRACING) {
			StringBuilder sb = new StringBuilder();
			sb.append("Processor ");
			sb.append(this.getClass().getName());
			sb.append(" was constructed for annotations:");
			for (AnnotationTypeDeclaration atd : _atds) {
				sb.append("\n");
				sb.append(atd.getQualifiedName());
			}
			APFactory.trace(sb.toString());
		}
	}

	public void process() {
		
		// *** Here we do element value error checking ***
		Messager messager = _env.getMessager();
		
		// obtain the declaration of the annotation we want to process
		AnnotationTypeDeclaration annoDecl = (AnnotationTypeDeclaration)_env.getTypeDeclaration(annotationNameWebMethod);
		
		// Don't trust anything you get from the mirror APIs; this runs as the user is typing
		if (annoDecl == null) {
			return;							
		}
		
		// get the annotated declarations
		Collection<Declaration> annotatedDecls = _env.getDeclarationsAnnotatedWith(annoDecl);
		if (annotatedDecls == null) {
			return;
		}
		
		// for each annotated declaration
		for (Declaration decl : annotatedDecls) {
			
			Collection<AnnotationMirror> mirrors = decl.getAnnotationMirrors();
			
			// for each annotation found, get a map of element name/value pairs
			for (AnnotationMirror mirror : mirrors) {
				// for the @WebMethod annotation
				if (mirror.getAnnotationType().getDeclaration().getQualifiedName().equals(annotationNameWebMethod)) {
					processWebMethod(messager, mirror, decl);
				}
				
			}
		}
	}

	// process a declaration annotated with @WebMethod
	private void processWebMethod(Messager messager, AnnotationMirror mirror, Declaration decl) {
		if (mirror == null || decl == null) {
			return;
		}
		
		// Get the set of element name/value pairs.
 		Map<AnnotationTypeElementDeclaration, AnnotationValue> valueMap = mirror.getElementValues();
		Set<Map.Entry<AnnotationTypeElementDeclaration, AnnotationValue>> valueSet = valueMap.entrySet();
		
		// iterate over the elements in valueSet, finding (operationName = ) elements
		// test (operationName = ) element for a value that begins with an alphabetic character
		for (Map.Entry<AnnotationTypeElementDeclaration, AnnotationValue> annoKeyValue : valueSet) {					
			AnnotationValue annoValue = annoKeyValue.getValue();	// get the value for this element

			// if this is element operationName=<string>
			if (annoKeyValue.getKey().getSimpleName().equals(elementNameOperationName)) {
				// TODO: extract the value from the annoKeyValue
				
				String elementValue = null; // TODO: get the value as a string
				
				// using regex to make sure the first character of the string is alphabetic
				Matcher m = alphaMatch.matcher(elementValue.substring(0, 1));
				if (!m.matches()) {
					messager.printError(null /* ??? get position of the value */, "operationName must begin with alpha character");
				}
			}

		}
		
		// Would the mirror deliver a non-method declaration that's annotated with @WebMethod? Oh yes.
		// The compiler will flag the error for the user, but that won't help this code.
		if (!(decl instanceof MethodDeclaration)) {
			return;
		}	
		
		// Do a check that for each private field on a complex type used as an input
		// parameter to a method annotated with @WebMethod, there exists a corresponding public getter.
		MethodDeclaration methodDecl = (MethodDeclaration) decl;
		// TODO: get the parameters
		
		// TODO: for each parameter
			// TODO: if the parameter is a Class type
				// TODO: get the ClassDeclaration
				// TODO: From it, obtain a list of the private fields, and a list of the public methods
				// Hint: to obtain the private fields, you will use DeclarationFilter.FILTER_PRIVATE
				// 		 to obtain the public getters, use the corresponding FILTER_PUBIC, in combination
				//		 with this anonymous one:
					DeclarationFilter getterFilter = new DeclarationFilter() {
						public boolean matches(Declaration d)
						{
							return (d.getSimpleName().substring(0, 3).equals("get"));
						}
					};
				// TODO: Crude test for this exercise: Make sure that there are at least as many public getXXX methods as there are private fields.
				// TODO: if not, use the messager to complain about the parameter.
			// TODO: if the parameter is a Primitive type
				// TODO: do nothing
			// TODO: else complain that only Class and Primitive types are supported in web methods.
		

	}
		
		


}
